"""
Rational RBF approximation of insviscid Burgers' equation
     u_t + 0.5(u^2)_x = 0   
  the solution forms a shock at time t = 1/pi
  
reference: S. A. Sarra and Y. Bai. A rational radial basis function method for accurately 
           resolving discontinuities and steep gradients. Applied Numerical Mathematics 130, 
           p. 131-142, 2018.  DOI: 10.1016/j.apnum.2018.04.001 
  
usage: run inviscidBurgersRRBF.py

"""


from numpy import linspace, inf, sin, zeros, pi
import matplotlib.pyplot as plt
from numpy.linalg import norm
import AnimatePlot as ap
from iqx import *
import rbfRational as rat
import rbfCenters as rc
import rbfMisc as rm
from scipy.optimize import brentq

phi = iqx()
s = 6.0
N = 100
k = 0.005
P = 1           # plot every P time step
finalT = 0.75
x = rc.xKT(N,-1,1,0.999)
r = phi.distanceMatrix1d(x)
B = phi.rbf(r,s)
H, H2 = phi.D1(r,s,r), phi.D2(r,s,r)
L, Bi = rat.rbfRationalCoreMultipleSetUp(B)

def F(u,t):
	u[0], u[-1] = 0.0, 0.0
	ra1, _ = rat.diff1dMultiple(0.5*u**2,B,Bi,L,H,H2)
	return -ra1

def exact(X,T):
	shockTime = 1/pi
	def f(x): return X - x + T*sin(pi*x)
	if X >= 0:                # after t=1/pi, for some X there will be more
		x0 = brentq(f,0,2)    # than one zero, find the correct one
	else:
		x0 = brentq(f,-2,0)
	return -sin(pi*x0)
	
	
	
def makePlot(): 
	plt.plot(x, u, 'b')
	plt.ylim([-1.2,1.2])
	
	
u = sin(pi*(x+1))
n, t = 0, 0
while n*k < finalT:
	u = rm.rk4(u,n*k,k,F)
	n += 1
	t += k
	u[0], u[-1] = 0.0, 0.0
	if n%P==0: ap.animate(makePlot)
	
	
ex = zeros((N,))
for j in range(N): ex[j] = exact(x[j],t)
plt.plot(x, ex,'r--')
plt.show()
	
print('error: {:1.2e}'.format( norm( u - ex, inf)))




