"""
Examines eigenvalue (ew) stability of a modified RBF PU method for the 2d diffusion 
equation with zero Dirichlet eigenvalues on a circular domain.

Three options are available for improving ew stability: MDI regularization,
a inverse system matrix filter, and Laplacian power artifical viscosity.  One
or more approach may be used at the same time.

usage: run heatCircleSarraPuEws.py

note: max real ew: 0.0 with only MDI
          
"""

import matplotlib.pyplot as plt
from numpy import arange, eye
import rbfPU as PU
import rbfCenters as rc
import rbfMisc as rm

from rbfClassX import rbfx
from gax import *

import warnings
warnings.filterwarnings('ignore')

plt.close('all')

phi = gax()

MDI = True   
FILTER = False
LAP = True
SAFE = False   # LU or Cholesky

mu, nuG, nuL = 0, 0, 0
order = 2   # Laplacian order

# --------------- centers -------------------------------------

N, Nb = 6000, 350
bpi = arange(Nb)
xi, yi = rc.circleCenters(N,cluster=False,ch=3,R=1,plt=False)
x, y = rc.circleBndCenters(xi,yi,1,Nb,tol=0.0125,plt=False)
N = len(x)
x += 1
y += 1

print('boundary centers: {:d}'.format(Nb))
print('   total centers: {:d}'.format(len(x)))

if MDI:           # MDI regularization parameter
	mu = 5e-15           
if FILTER:        # hyperviscosity - inverse system matrix filter
	nuG = 8e-12     
if LAP:           # hyperviscosity - Laplacian power
	nuL = 7e-14


print('            MDI mu: {:1.2e}'.format(mu))
print(' inverse system nu: {:1.2e}'.format(nuG))
print('Laplacian power nu: {:1.2e}'.format(nuL))


# ---------- modified PU setup ----------------------------------

radius, xc, yc, st = PU.cover2d(x,y,5,0.75,0)
X = list( zip(x,y) )
shape = PU.patchShape(X,st,phi,mu=0,s0=0.5,ds=0.01)

from W32 import *
phiC = W32()

n = len(st)                                     # number of patches
w = PU.patchWeights(st,xc,yc,x,y,radius,phiC)

d = []
for k in range(n):
	st0 = st[k]
	s = shape[k]
	r, rx, ry = rbfx.distanceMatrix2d(x[st0],y[st0])
	B, H = phi.rbf(r,s), phi.L(r,s)
	Hv = phi.hyperViscosity(2,order,s,r)
	n0 = B.shape[0]
	I = eye(n0)
	d.append( phi.dm( B, H - nuG*I - nuL*Hv, mu, False ) )
	
D = PU.globalDMSarra(N,st,w,d)
D[bpi,:] = 0      # zero Dirichlet BCs

rm.dmEWs(D,dt=0.00015,p1=True,p2=True,verbose1=True,verbose2=True)
