"""    Python Radial Basis Function Toolkit (PRBFT)

Project homepage:    http://www.scottsarra.org/rbf/rbf.html
Contact e-mail:      sarra@marshall.edu

Copyright (c) 2023 Scott A. Sarra

Licensing: PRBFT is under the GNU General Public License ("GPL").

GNU General Public License ("GPL") copyright permissions statement:
**************************************************************************
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


The module rbfClassX contains the class rbfx.
"""

import scipy as sp
import scipy.optimize as so
import scipy.linalg as la
from abc import ABC, abstractmethod
from numpy.random import rand         

class rbfx(ABC):
	"""
	The base class for positive definite RBFs that are infinitely differentiable
	and that have a shape parameter.
	
	usage: from rbfClassX import rbfx
	
	"""

		
#----------------------------------------------------------------------------
#----------------------- Abstract methods------------------------------------
#            used to define a common interface for all subclasses
#                   must be implemented by all subclasses.
#----------------------------------------------------------------------------

# ------------------------------------------------------------------------------		
# --------------------- rbf definition -----------------------------------------
# ------------------------------------------------------------------------------
		
	@abstractmethod
	def rbf(self,r,s):
		"""
		This functions defines the RBF.
		
		inputs:
		  r    distance matrix
		  s    shape parameter
		  
		output:
		  d    the RBF evaluated at r
		"""
		pass
	
# ------------------------------------------------------------------------------	
# ---------------------- derivatives -------------------------------------------
# ------------------------------------------------------------------------------

	@abstractmethod
	def D1(self,r,s,x): 
		"""
		First partial derivative of the basis function wrt x in any space dimension.
		
		inputs:
			r    distance matrix
			s    shape parameter
			x    distance matrix of the variable that the derivative is with respect to,
				 for example rx, ry, rz, etc.
		output:
			d    first derivative of the basis function wrt to x
		
		"""
		pass
	

		
	@abstractmethod
	def D2(self,r,s,x):
		"""
		Second partial derivative of the basis function wrt x in any space dimension.
		
		inputs:
			r    distance matrix
			s    shape parameter
			x    distance matrix of the variable that the derivative is with respect to,
				 for example rx, ry, rz, etc.
		output:
			d    second derivative of the basis function wrt to x
		
		"""
		pass
	
	@abstractmethod
	def D3(self,r,s,x):
		"""
		Third partial derivative of the basis function wrt x in any space dimension.
		
		inputs:
			r    distance matrix
			s    shape parameter
			x    distance matrix of the variable that the derivative is with respect to,
				 for example rx, ry, rz, etc.
		output:
			d    third derivative of the basis function wrt to x
		
		"""
		pass

	@abstractmethod
	def D4(self,r,s,x):
		"""
		Fourth partial derivative of the basis function wrt x in any space dimension.
		
		inputs:
			r    distance matrix
			s    shape parameter
			x    distance matrix of the variable that the derivative is with respect to,
				 for example rx, ry, rz, etc.
		output:
			d    fourth derivative of the basis function wrt to x
		
		"""
		pass
	

	
	@abstractmethod
	def div(self,r,s,x,y=0,z=0):
		"""
		divergence of the basis function in 1, 2, and 3 space dimensions, for example in 3 dimensions
		    d = u_x + u_y + u_z
		
		inputs:
			r    distance matrix
			s    shape parameter
			x    distance matrix wrt x
			y    distance matrix wrt y
			z    distance matrix wrt z
	
		output:
			d    divergence of the basis function
		
		"""
		pass
	
	
	@abstractmethod
	def L(self,r,s):
		"""
		Laplacian of the basis function in 2 space dimensions
		    d = u_{xx} + u_{yy}
		
		inputs:
			r    distance matrix
			s    shape parameter
	
		output:
			d    Laplacian of the basis function
		
		"""
		pass
	
	@abstractmethod
	def B(self,r,s,x,y):
		"""
		Biharmonic operator applied to the basis function in 2 space dimensions
		    d = u_{xxxx} + 2 u_{xxyy} + u_{yyyy}
		
		inputs:
			r    distance matrix
			s    shape parameter
	
		output:
			d    Biharmonic operator applied to the basis function
		
		"""
		pass
		
	
	@abstractmethod       
	def hyperViscosity(self,dim,order,shape,r):
		"""
		*** NOTE: currently only implemented for the GA ***
		
		Artifical viscosity via a power of the Laplacian
		
		inputs:
		   dim      the space dimension
		   order    the power of the Laplacian
		   shape    the shape parameter
		     r      distance matrix
			 
		output:
		   d        power of the Laplacian operator applied to the basis funcctions
		"""
		pass      
	

# ------------------------------------------------------------------------------
# -------------------- distance matrices ---------------------------------------
# ------------------------------------------------------------------------------


	@staticmethod
	def distanceMatrix(xc, x=None, nargout = 1):
		"""
		Forms distances matrix in 1 to 3 space dimensions
		
		inputs xc and x should be of the form, for example in 2d, X = list( zip(x,y) )
		"""

		xc = sp.array(xc)
		try:    # in 1d xc may not have a second dimension
			s = sp.shape(xc)[1]
		except IndexError:
			s = 1

		if s == 1:
			if x is not None:
				x = sp.array(x)
				r = rbfx.distanceMatrix1d(xc,x)
			else:
				r = rbfx.distanceMatrix1d(xc)
			return r
		elif s == 2:
			if x is not None:
				x = sp.array(x)
				r, rx, ry = rbfx.distanceMatrix2d(xc[:,0], xc[:,1], x[:,0], x[:,1])
			else:
				r, rx, ry = rbfx.distanceMatrix2d(xc[:,0], xc[:,1])
			if nargout == 1:
				return r
			else:
				return r, rx, ry
		elif s == 3:
			if x is not None:
				x = sp.array(x)
				r, rx, ry, rz = rbfx.distanceMatrix3d(xc[:,0], xc[:,1], xc[:,2], x[:,0], x[:,1], x[:,2])
			else:
				r, rx, ry, rz = rbfx.distanceMatrix3d(xc[:,0], xc[:,1], xc[:,2])
			if nargout == 1:
				return r
			else:
				return r, rx, ry, rz
		

# ------------------------------------------------------------------------------

	@staticmethod
	def distanceMatrix1d(xc, x=None):   # returns the signed distance matrix
		""" distanceMatrix1d
		inputs
		  xc   N x 1 vector of centers
		  x    M x 1 vector of evaluation points (optional)
		
		outputs
		  r    signed distance matrix
		       N x N,   r_{ij} = dist between center i and j if called as distanceMatrix1d(xc)
		       M x N,   r_{ij} = dist between evaluation point i and center j if called as distanceMatrix1d(xc,x)
		       N x N/2  if called as r = phi.distanceMatrix1d(xc(1:N/2),xc)
		                returns the left half of the distance matrix needed
		                for a centro center distribution
		
		example usage:
		""" 
		
		xc = sp.reshape(xc,(len(xc),1))
		if x is None:            
			return xc - xc.T
		else:
			x = sp.reshape(x,(len(x),1))
			return x - xc.T

# -------------------------------------------------------------------------------
		
	@staticmethod
	def distanceMatrix2d(xc, yc, x=None, y=None):
		"""  distanceMatrix2d

			inputs
			  xc   N x 1 vectors of centers  XC = (xc,yc)
			  yc
			  x    M x 1 vectors of evaluation points X = (x,y) (optional)
			  y
			
			outputs
			  r    signed distance matrix
			       N x N,   r_{ij} = dist between center i and j if called as distanceMatrix1d(xc,yc)
			       M x N,   r_{ij} = dist between evaluation point i and center j if called as distanceMatrix1d(xc,yc,x,y)
			       N x N/2  if called as r = phi.distanceMatrix1d(xc(1:N/2),yc(1:N/2),xc,yc)
			                returns the left half of the distance matrix needed
			                for a centro center distribution
			
			example usage:
		"""
		
		xc, yc = sp.reshape(xc,(len(xc),1)), sp.reshape(yc,(len(yc),1))
		if x is None: 
			rx, ry = xc - xc.T, yc - yc.T
		else:
			x, y = sp.reshape(x,(len(x),1)), sp.reshape(y,(len(y),1))
			rx, ry = x - xc.T, y - yc.T

		r = sp.sqrt( rx**2 + ry**2 )	
		return r, rx, ry
		
# -------------------------------------------------------------------------------
		
	@staticmethod
	def distanceMatrix3d(xc, yc, zc, x=None, y=None, z=None):
		""" distanceMatrix3d

		inputs
		  xc   N x 1 vectors of centers  XC = (xc,yc,zc)
		  yc
		  zc
		  x    M x 1 vectors of evaluation points X = (x,y,z) (optional)
		  y
		  z
		
		outputs
		  r    signed distance matrix
		       N x N,   r_{ij} = dist between center i and j if called as distanceMatrix1d(xc,yc,zc)
		       M x N,   r_{ij} = dist between evaluation point i and center j if called as distanceMatrix1d(xc,yc,zc,x,y,z)
		       N x N/2  if called as r = phi.distanceMatrix1d(xc(1:N/2),yc(1:N/2),zc(1:N/2),xc,yc,zc)
		                returns the left half of the distance matrix needed
		                for a centro center distribution
		
		example usage: 1) 

		"""
		
		xc, yc, zc = sp.reshape(xc,(len(xc),1)), sp.reshape(yc,(len(yc),1)), sp.reshape(zc,(len(zc),1))
		if x is None: 
			rx, ry, rz = xc - xc.T, yc - yc.T, zc - zc.T
		else:
			x, y, z = sp.reshape(x,(len(x),1)), sp.reshape(y,(len(y),1)), sp.reshape(z,(len(z),1))
			rx, ry, rz = x - xc.T, y - yc.T, z - zc.T
			
		r = sp.sqrt( rx**2 + ry**2 + rz**2 )
		return r, rx, ry, rz

		
# ----------------------------------------------------------------------------
# ---------- regularized SPD linear system solvers ---------------------------
# ----------------------------------------------------------------------------

	@staticmethod
	def solve(B, f, mu=5e-15, safe=False):
		""" solves the SPD linear system B a = f for a with the option to regularize
		        by the method of diagonal increments (MDI)
		inputs 
		   B    N x N symmetric positive definite (SPD) matrix
		   f    N x 1 vector
		  mu    (optional) MDI regularization parameter.  Use mu = 0 for no regularization
		 safe   True - uses LU factorization
		        False - uses a Cholesky factorization.  Faster, but it the matrix is 
		                severely ill-conditioned and/or the regularization parameter is too
		                small the matrix may fail to be numerically SPD and the Cholesky
		                factorization will fail
		
		outputs
		   a    N x 1 solution vector
		
		example usage:
		  1) /rbfX/examples/papers/JAMP2023/mdiRegularization.py
		"""
		
		if mu>0: B[sp.diag_indices_from(B)] += mu    # MDI regularization
			
		if safe:    # LU with partial pivoting
			lu, piv = la.lu_factor(B)
			return la.lu_solve((lu, piv), f)
		else:       # Cholesky
			c, low = la.cho_factor(B)
			return la.cho_solve((c, low), f)
			
			
# ----------------------------------------------------------------------------

	@staticmethod
	def dm(B, H, mu=5e-15, safe=False):  
		"""  forms the deriviative matrix D = H*inv(B) by solving the system D B = H for D
		inputs:rbfCompactClass
		   B   N x N SPD system matrix
		   H   N x N derivative evaluation matrix
		  mu    (optional) MDI regularization parameter.  Use mu = 0 for no regularization
		 safe   True - uses LU factorization with partial pivoting
		        False - uses a Cholesky factorization.  Faster, but if the matrix is 
		                severely ill-conditioned and/or the regularization parameter is too
		                small the matrix may fail to be numerically SPD and the Cholesky
		                factorization will fail
		
		output:
		   D   N x N differentiation matrix
		
		example usage:
		 1) 	
		"""
			  
		if mu>0: B[sp.diag_indices_from(B)] += mu   # MDI regularization
		if safe:
			lu, piv = la.lu_factor(B.T)
			D = la.lu_solve((lu, piv), H.T)
		else:    
			c, low = la.cho_factor(B.T)              # Cholesky
			D = la.cho_solve((c, low), H.T)
		return D.T
	
# ----------------------------------------------------------------------------
# ---------- alternative SPD linear system solvers ---------------------------
# ----------------------------------------------------------------------------



	@staticmethod
	def solveAlt(B, f, mu=5e-15):
		""" Solves a theoretically SPD linear system B a = f for a via a Cholesky
		factorization.  If B is severely ill-conditioned and B ceases to be numerically
		SPD LU factorization is used.
		
		inputs 
		   B    N x N symmetric positive definite (SPD) matrix
		   f    N x 1 vector
		  mu    (optional) MDI regularization parameter.  Use mu = 0 for no regularization
	
		outputs
		   a       N x 1 solution vector
		   flag    0 Cholesky , 1 LU
		
		example usage:
		   1) /examples/papers/JAMP2023/mdiRegularization.py
		"""
		flag = 0
		if mu>0: B[sp.diag_indices_from(B)] += mu    # MDI regularization
		
		try:
			c, low = la.cho_factor(B)
			a = la.cho_solve((c, low), f)
		except la.LinAlgError:
			lu, piv = la.lu_factor(B)
			a = la.lu_solve((lu, piv), f)
			flag = 1
			
		return a, flag
			
		
		
	
# ---------------------------------------------------------------------------
# ---------- constant shape parameter selection -----------------------------
# ---------------------------------------------------------------------------

	@staticmethod
	def spdShape(phi, x, s0, ds=0.02, mu=0.0, verbose=False):
		"""  finds smallest shape parameter within distance ds in which a system matrix is SPD
		
		inputs:
		  phi   SPD RBF
		    x   centers, in the format X = list( zip(x,y) )
		   s0   initial shape parameter	
	       mu  regularization parameter for shape selection, recommended mu = 0 (default)
	  verbose  logical variable for printing info
		
		output:
		   s   shape parameter
		   L   (regularized, mu>0) Cholesky factorization of the RBF system matrix. 
		
		example usage:
		 1)
		 
		 NOTE: It is best to find the "optimal" shape using mu = zero and then use mu>0 to
              factorize the system matrix with the optimal shape
		"""

		r = rbfx.distanceMatrix(x)
		s = s0
		
		try:
			B = phi.rbf(r,s)
			B[sp.diag_indices_from(B)] += mu
			L = la.cho_factor(B)
		except la.LinAlgError:
			SPD = False
			if verbose: print('initially not SPD at s0 = {:1.2f}'.format(s0))
		else:
			SPD = True
			if verbose: print('initially SPD at s0 = {:1.2f}'.format(s0))
				
		if not SPD:
			while not SPD:
				try:
					B = phi.rbf(r,s)
					B[sp.diag_indices_from(B)] += mu
					L = la.cho_factor(B)
				except la.LinAlgError:
					s += ds
				else:
					SPD = True
					if verbose: print('becomes SPD at shape parameter: {:f}'.format(s))
			return s, L
					
		if SPD:
			while SPD:
				try:
					B = phi.rbf(r,s)
					B[sp.diag_indices_from(B)] += mu
					L = la.cho_factor(B)
				except la.LinAlgError:
					SPD = False
					if verbose: print('ceases to be SPD at shape parameter: {:f}'.format(s))
				else:
					s -= ds			
			return s, L




# ----------------------------------------------------------------------------   
# -------------- variable shape parameters -----------------------------------
# ----------------------------------------------------------------------------



	@staticmethod
	def variableShape(sMin,sMax,opt,N,M=None):
		"""
		inputs
		  sMin    minimum value of the shape parameter
		  sMax    maximum value of the shape parameter
		   N      number of columns
		   M      number of rows
		 opt  1, exponentially varying (Kansa)
				 Computers and Mathematics with Applications v. 24, no. 12, 1992. 
			  2, linearly varying
			  3, randonly varying (Sarra and Sturgil)
				 Engineering Analysis with Boundary Elements, v. 33, p. 1239-1245, 2009.
		
		 outputs
		   s1   N x N matrix with constant shapes in each column
		   s2   M x N matrix with constant shapes in each column 
				(for interpolation evaluation matrix). Only returned if
				M is not equal to None

		example usage:
		  1) 
		"""
		
		if opt==1:
			sMin, sMax = sMin**2, sMax**2
			s = sp.sqrt( sMin*(sMax/sMin)**(sp.arange(N)/(N-1)) )
		elif opt==2:
			s = sMin + ((sMax - sMin)/(N-1))*sp.arange(N)
		else:
			s = sMin + (sMax - sMin)*rand(1,N)
			
		if M is not None:
			return sp.tile(s,(N,1)), sp.tile(s,(M,1))
		else:
			return sp.tile(s,(N,1))


# ----------------------------------------------------------------------------   
# -------------- miscellaneous------------------------------------------------
# ----------------------------------------------------------------------------
	
	@staticmethod		
	def fillSeparationDistance(x,y=None,z=None):
		""" Calculates the fill distance and separation distance for a set of
		centers located in 1d, 2d, or 3d.
		
		input:
			x, y, z    center locations
			
		output:
			h          fill distance
			q          minimun separation distance
		"""
		
		N = len(x)
		if y==None and z==None:
			r = abs( rbfx.distanceMatrix1d(x) )
		elif z==None:
			r, rx, ry = rbfx.distanceMatrix2d(x,y)
		else: 
			r, rx, ry, rz = rbfx.distanceMatrix3d(x,y,z)
			
		temp = sp.amin(r + 100*sp.eye(N),axis=0)
		return sp.amax(temp), 0.5*sp.amin(temp)


# ----------------------------------------------------------------------------


	@staticmethod		
	def kappa(B,mu=0):
		"""
		Calculates the 2-norm condition number of a symmetric system matrix and
		the matrix regularized via MDI.
		
		input:
		  B     symmetric system matrix
		  mu    MDI regularization parameter
		
		example usage: condNumBench.py
		"""
		ew = la.eigvalsh(B,check_finite=False)
		kappaB, kappaBMu = ew[-1]/ew[0], (ew[-1] + mu)/(ew[0] + mu)
		return kappaB, kappBMu




# ----------------------------------------------------------------------------


