"""
Module for implementing local RBF methods.

import suggestion: import rbfLocal as LR

------------------------- function summary --------------------------------------

stencils - efficiently constructs stencils using a KD tree
shapes - NSPD shape parameter selection
weights - calculates the weights for each stencil
plotStencils - visualization of centers and stencils

"""

import scipy as sp
import matplotlib.pyplot as pplt
from scipy.sparse import lil_matrix
from scipy import sparse
from scipy.linalg import LinAlgError, cho_factor, cho_solve
from scipy import spatial
from rbfClassX import rbfx
import warnings

# --------------------------------------------------------------------------------
# ------------- construct local stencils -----------------------------------------
# --------------------------------------------------------------------------------


def stencils(X,ns):
    """ Constructs local RBF stencial using a KD Tree.
    
    Inputs:
      X   x and y coordinates of the centers in the format X = list( zip(x,y) )
     ns   stencil size
    Output:
     st   indices of the centers in the stencil for each of the N centers
    """
    tree = spatial.cKDTree(X)
    q = tree.query(X,ns) 
    return q[1]


# ---------------------------------------------------------------------------------
# ------- find shape parameters for each stencil ----------------------------------
# ---------------------------------------------------------------------------------

def shapes(X,st,phi,mu=0,s0=0.5,ds=0.01,verbose=False):
    """ Shape parameter selection for each local RBF stencils based on the local system 
        matrix being numerical positive definite.
        
    Inputs:
       X     x and y coordinates of the centers in the format X = list( zip(x,y) )
      st     indices of the centers in the stencil for each of the N centers
      phi    a SPD RBF
       mu    method of diagonal increments regularization parameter
       s0    initial shape for the first stencil
       ds    increment to change shape value
   verbose   print information for each shape search
   
   Outputs:
       S     shape parameter for each of the N stencils
    """
    
    sh = sp.shape(X)
    N, d = sh[0], sh[1]          # number of centers, space dimension
    
    initS = s0
    S = sp.zeros(N)              # shape parameter on each stencil
    
    if d==1:
        x = sp.array(X)
        x = sp.reshape(x,(N,))
        for k in range(N):
            x0 = x[st[k]]
            S[k], _ = rbfx.spdShape(phi,x0,initS,ds,mu,verbose)
            initS = S[k]        # initial shape on next stencil = final shape on previous stencil
            
    elif d==2:
        x, y = zip(*X)
        x, y = sp.array(x), sp.array(y)
        for k in range(N):
            x0 = list( zip(x[st[k]],y[st[k]]) )
            S[k], _ = rbfx.spdShape(phi,x0,initS,ds,mu,verbose)
            initS = S[k]
        
    elif d==3:
        x, y, z = zip(*X)
        x, y, z = sp.array(x), sp.array(y), sp.array(z)
        for k in range(N):
            x0 = list( zip(x[st[k]],y[st[k]],z[st[k]]) )
            S[k], _ = rbfx.spdShape(phi,x0,initS,ds,mu,verbose)
            initS = S[k]
               
    return S


# -----------------------------------------------------------------------------
# --- weights for linear differential operators on each stencil ---------------
# -----------------------------------------------------------------------------


def weights(X,st,phi,s,L,mu=0,visc=0,visc2=0,dim=2,order=1):
    """ Calculates the weights for each local RBF stencil.
    
    Inputs:
     X    centers in the format X = list( zip(x,y) )
    st    stencil for each center
   phi    a SPD RBF
     s    array with shape parameter for each stencil
     L    a linear differential operator
    mu   MDI regularization parameter
  visc    hyperviscosity parameter - inv(B) regularization

  visc2   hyperviscosity parameter - Laplacian power
  dim     space dimension: 1,2,or 3
 order    Laplacian order

   Outputs:
     D   sparse differentiation matrix that discretizes L
    """
    sh = sp.shape(X)
    N, d, ns = sh[0], sh[1], len(st[0,:])  # number of centers, space dimension, stencil size
   
    o = sp.ones(ns)
    Ins = sp.eye(ns)
    ov = sp.zeros(ns)
    ov[0] = 1.0     #  ov = (1,0,0,...)
    row, col, val = [], [], []
    warnings.filterwarnings('ignore')    # ignore ill-conditioned matrix warnings
    
    if d==1:
        x = sp.array(X)
        x = sp.reshape(x,(N,))
    elif d==2:
        x, y = zip(*X)
        x, y = sp.array(x), sp.array(y)
    elif d==3:
        x, y, z = zip(*X)
        x, y, z = sp.array(x), sp.array(y), sp.array(z)
    
    for k in range(N):  # calculate weights on each stencil
        st0 = st[k]
        
        if d==1:
            r = rbfx.distanceMatrix1d(x[st0])
        elif d==2:
            r, rx, ry = rbfx.distanceMatrix2d(x[st0],y[st0])              
        elif d==3:
            r, rx, ry, rz = rbfx.distanceMatrix3d(x[st0],y[st0],z[st0])              

        B = phi.rbf(r,s[k])
   
        row.append( (k*sp.ones(ns,dtype=int)).tolist())   # row k
        col.append(st0.tolist())
        
        if d==1:
            xs = x[k]- x[st0]                           
            rL = abs( xs )
            h = L(rL,xs,s[k],phi)
        elif d==2:
            xs, ys = x[k]- x[st0], y[k]- y[st0]                           
            rL = sp.sqrt( xs**2 + ys**2 )
            h = L(rL,xs,ys,s[k],phi)
        elif d==3:
            xs, ys, zs = x[k]- x[st0], y[k]- y[st0], z[k]- z[st0]                            
            rL = sp.sqrt( xs**2 + ys**2 + zs**2 )
            h = L(rL,xs,ys,zs,s[k],phi)
            
        if visc>0:
            h -= visc*ov
        if visc2>0:
            h -= visc2*ov*phi.hyperViscosity(dim,order,s[k],rL)  
            
        w = rbfx.solve(B.T, h.T, mu, False)
#        w = rbfx.solve(B.T, (h - visc*ov).T, mu, False)
        val.append(( w.T ).tolist()) 
        
    row = [item for sublist in row for item in sublist]	
    col =  [item for sublist in col for item in sublist]
    val = [item for sublist in val for item in sublist]

    D = sparse.coo_matrix((val,(row,col)),shape=(N,N)).tocsr()    
    
    return D



# --------------------------------------------------------------------------------
# -------------- visualization ---------------------------------------------------
# --------------------------------------------------------------------------------

def plotStencil(x,y,centerNumber,st):
    """ Plot the stencil for one center.
    
    Inputs:
       x, y           center coordinates
     centerNunber     plot the stencil for this center
       st             the stencils for each center (the output of rbfLocal.stencils)
    """
    pplt.scatter(x,y,c='b',marker='.')
    pplt.scatter(x[st[centerNumber]],y[st[centerNumber]],c='k',marker='.')
    pplt.scatter(x[centerNumber],y[centerNumber],c='r',marker='*')
    pplt.show()


# ------------------------------------------------------------------------------------
