"""
Module for working with rational RBF methods.

Reference:  S. A. Sarra and Y. Bai. "A rational radial basis function method
for accurately resolving discontinuities and steep gradients".
Applied Numerical Mathematics 130, p. 131-142, 2018.
DOI: 10.1016/j.apnum.2018.04.001
			
import suggestion: import rbfRational as rat

-------------- function summary -------------------------------------

rbfRationalCoreOne - Called to setup single approximations involving a
                     set of centers and a shape parameter.  Returns the 
                     expansion coefficients of p and q to be used in
                     interpolation and differentiation. The following
                     functions call this method: interpolate1d,
                     interpolate2d, diff1dOne, and diff2dOne


rbfRationalCoreMultipleSetUp - Sets up for multiple approxmations
                     involving the same centers and shape parameter.
                     Returns the inverse and Cholesky factorization of B.

rbfRationalCoreMultiple - Uses the inverse and Cholesky factorization of
                     B from rbfRationalCoreMultipleSetUp to calculate the
                     expansion coefficients of p and q.

interpolate1d - Uses the expansion coefficients of p and q from rbfRationalCoreOne
                to interpolate a function of one variable. 

interpolate2d - Uses the expansion coefficients of p and q from rbfRationalCoreOne
                to interpolate a function of two variables. 

diff1dOne - Uses the expansion coefficients of p and q from rbfRationalCoreOne
            to differentiate a function of one variable. 

diff1dMultiple - Passes the previously calculated inverse and Cholesky
                 factorization of B to rbfRationalCoreMultiple to get new
                 expansion coefficients of p and q of to be used to
                 differentiate functions of one variable using the same 
                 centers and shape parameter.  For discretizing space 
                 derivatives of time-dependent PDEs.

diff2dOne - Uses the expansion coefficients of p and q from rbfRationalCoreOne
            to differentiate a function of two variables. 

diff2dMultiple - same as diff1dMultiple except it is for a function of 2 variables.

Note, May 2023.  In some cases the Python version of the rational method
  does not perform as well as the Matlab version.  Possibly due to the
  eigensolver used.  This will be addressed in a future release.
  

"""

import scipy as sp
from scipy.linalg import cho_factor, cho_solve, eigh, eig
from scipy.sparse.linalg import eigs
from rbfClassX import rbfx




def rbfRationalCoreOne(B,fc,mu=0):
	""" 
	input
	   B    system matrix
	  fc    function values at the center locations
	  mu    optional MDI regularization parameter
	
	output
	 pAlpha  standard RBF exapansion coefficients of the vector p 
	 qAlpha  standard RBF exapansion coefficients of the vector q
		
	called by: interpolate1d, interpolate2d, rbfRational.diff1dOne 
           and rbfRational.diff2dOne
	"""
	N = sp.shape(B)[0]
	if mu>0: B[sp.diag_indices_from(B)] += mu    # MDI regularization
	L = cho_factor(B)
	Bi = cho_solve(L,sp.eye(N))     # B inverse
	Df = sp.diag(fc)
	K = 1.0/sum(fc**2)
	S = sp.diag(1.0/(K*fc**2 + 1))@( K*Df@Bi@Df + Bi )

	ew, ev = eig(S)
	#ew, ev = eigh(S)
	I = sp.argmin(ew)
	ev = ev.T
	q = ev[I]

	p = Df@q
	pAlpha, qAlpha = cho_solve(L,p), cho_solve(L,q)
	return pAlpha, qAlpha

# --------------------------------------------------------------

def interpolate1d(phi,s,fc,xc,x,mu=0):
	""" 1d rational RBF interpolation

	inputs
	  phi   a RBF (a subclass of rbfx such as iqx or gax)
		s   shape parameter
	   fc   function values at centers xc
	   xc   centers
		x   evaluation points
	   mu   optional MDI regularization parameter
	
	output
	   fa   approximate function values at evaluation points x
	
	example usage: stepFunctionRRBF.py
	"""
	
	r, re = rbfx.distanceMatrix1d(xc), rbfx.distanceMatrix1d(xc,x);
	B = phi.rbf(r,s)     # system matrix
	H = phi.rbf(re,s)    # evaluation matrix
	pAlpha, qAlpha = rbfRationalCoreOne(B,fc,mu)
	fa = (H@pAlpha)/(H@qAlpha) 
	return  fa

# --------------------------------------------------------------

def diff1dOne(phi,s,fc,xc,mu=0):   # DOES NOT WORK
	""" input
		phi   radial basis function, e.g. iqx, gax
		  s   shape parameter
		 fc   function value at center locations
		 xc   centers
		 mu   optional MDI regularization parameter
		 
		 output
		   fa1   1st derivative
		   fa2   2nd derivative
	 
	 example usage: diffRRBF.py
	"""
	r = rbfx.distanceMatrix1d(xc)   
	B = phi.rbf(r,s)     # system matrix
	H = phi.D1(r,s,r)    # 1st derivative evaluation matrix
	pAlpha, qAlpha = rbfRationalCoreOne(B,fc,mu)
	Bp, Bq = B@pAlpha, B@qAlpha
	Hp, Hq = H@pAlpha, H@qAlpha
	
	fa1 = ( Bq*Hp - Bp*Hq )/Bq**2
	
	H2 = phi.D2(r,s,r)
	Hpp, Hqq = H2@pAlpha, H2@qAlpha
	fa2 = ( 2*Bp*Hq**2 + Bq**2*Hpp - Bq*( 2*Hp*Hq + Bp*Hqq ) )/Bq**3
		
	return fa1, fa2
	
# --------------------------------------------------------------

def rbfRationalCoreMultipleSetUp(B,mu=0):
	"""  
	input
		B   system matrix
	   mu   optional MDI regularization parameter
	output
		L   Cholesky factorization of B
	   Bi   inverse of B
	
	example usage: inviscidBurgersRRBF.py, adv1dRRBF.py  

	"""
	N = sp.shape(B)[0]
	if mu>0: B[sp.diag_indices_from(B)] += mu    # MDI regularization
	L = cho_factor(B)
	Bi = cho_solve(L,sp.eye(N))
	return L, Bi

# --------------------

def rbfRationalCoreMultiple(B,Bi,L,fc):
	""" input
		B  system matrix
	   Bi  inverse of B (computed by rbfRationalCoreMultipleSetUp)
		L  Cholesky factorization of B (computed by rbfRationalCoreMultipleSetUp)
	   fc  function value at the center locations
	 
	 output
	  pAlpha   standard RBF expansion coefficients of the vector p
	  qAlpha   standard RBF expansion coefficients of the vector q
	 
	 called by: diff1dMultiple and diff2dMultiple
	"""
	Df = sp.diag(fc)
	K = 1.0/sum(fc**2)
	S = sp.diag(1.0/(K*fc**2 + 1))@( K*Df@Bi@Df + Bi )
	ew, ev = eig(S)
	I = sp.argmin(ew)
	ev = ev.T
	q = ev[I]
	p = Df@q
	pAlpha, qAlpha = cho_solve(L,p), cho_solve(L,q)
	return pAlpha, qAlpha
	
# --------------------

def diff1dMultiple(fc,B,Bi,L,H,H2):
	""" inputs:
	 fc   function values at centers xc
      B   system matrix
	 Bi   inverse of the system matrix
	  L   Cholesky factorization of the system matrix
	  H   first derivative evaluation matrix
	 H2   second derivative evaluation matrix

  output
	 fa1   approximate 1st derivative values 
	 fa2               2nd
  
  example usage: inviscidBurgersRRBF.py, adv1dRRBF.py 
	"""
	pAlpha, qAlpha = rbfRationalCoreMultiple(B,Bi,L,fc)
	Bp, Bq = B@pAlpha, B@qAlpha
	Hp, Hq = H@pAlpha, H@qAlpha
	
	fa1 = ( Bq*Hp - Bp*Hq )/Bq**2
	
	Hpp, Hqq = H2@pAlpha, H2@qAlpha
	fa2 = ( 2*Bp*Hq**2 + Bq**2*Hpp - Bq*( 2*Hp*Hq + Bp*Hqq ) )/Bq**3
		
	return fa1, fa2
	

# ------------------------ 2d -------------------------------------------

def interpolate2d(phi,s,fc,xc,x,yc,y,mu=0):
	""" inputs
	phi   a RBF (a subclass of rbfx such as iqx or gax)
	  s   shape parameter
	 fc   function values at centers xc
	 xc   centers
	 yc   
	  x   evaluation points
	  y   
	 mu   optional MDI regularization parameter
  
  output
	 fa   approximate function values at evaluation points x
  
  example usage: interp2dRRBF.py
	"""
	
	r, NA, NA = rbfx.distanceMatrix2d(xc,yc)
	re, NA, NA = rbfx.distanceMatrix2d(xc,yc,x,y);
	B = phi.rbf(r,s)     # system matrix
	H = phi.rbf(re,s)    # evaluation matrix
	pAlpha, qAlpha = rbfRationalCoreOne(B,fc,mu)
	fa = (H@pAlpha)/(H@qAlpha) 
	return  fa

# ------------------------------------------------------------------------------	
# -------- UNDER CONSTRUCTION --------------------------------------------------
# -------- TO BE FINSIHED IN VERSION 1.2.1 -------------------------------------
# ------------------------------------------------------------------------------

def diff2dOne(phi,s,fc,xc,yc,second=False,mu=0):     # NOT TESTED
	"""  1st and 2nd derivatives
	inputs
	  phi   a RBF (a subclass of rbfx such as iqx or gax)
		s   shape parameter
	   fc   function values at centers (xc,yc)
	second  logical variable, if true a second derivative is also returned
	   xc   centers
	   yc
	
	output
	   fx   approximate 1st derivative values at evaluation points x
	   fy
	   fxx  (optional)
	   fyy  (optional)
	"""
	
	r, rx, ry = rbfx.distanceMatrix2d(xc,yc)
	B = phi.rbf(r,s)                         # system matrix
	Hx, Hy = phi.D1(r,s,rx), phi.D1(r,s,ry)  # 1st derivative evaluation matrices
	pAlpha, qAlpha = rbfRationalCoreOne(B,fc,mu)
	Bp, Bq = B@pAlpha, B@qAlpha
	Hp, Hq = Hx@pAlpha, Hx@qAlpha
	
	fx = ( Bq*Hp - Bp*Hq )/Bq**2
	
	Hpy, Hqy = Hy@pAlpha, Hy@qAlpha
	fy = ( Bq*Hpy - Bp*Hqy )/Bq**2
	
	if second:
		H2x, H2y = phi.D2(r,s,rx), phi.D2(r,s,ry)    # 2nd derivative evaluation matrices
		Hpp, Hqq = H2x@pAlpha, H2x@qAlpha
		Hppy, Hqqy = H2y@pAlpha, H2y@qAlpha
		fxx = ( 2*Bp*Hq**2 + Bq**2*Hpp - Bq*( 2*Hp*Hq + Bp*Hqq ) )/Bq**3
		fyy = ( 2*Bp*Hqy**2 + Bq**2*Hppy - Bq*( 2*Hpy*Hqy + Bp*Hqqy ) )/Bq**3
		
	return fx, fy, fxx, fyy





def diff2dMultiple(fc,B,Bi,L,Hx,Hy,H2x,H2y,second=False,mu=0):
	"""  
	inputs:
	 fc   function values at centers xc
      B   system matrix
	 Bi   inverse of the system matrix
	  L   Cholesky factorization of the system matrix
	  Hx  first derivative evaluation matrix wrt x
	  Hy  first derivative evaluation matrix wrt y
	 H2x  second derivative evaluation matrix wrt x
     H2y  second derivative evaluation matrix wrt x
	
	output
	   fx   approximate 1st derivative values at evaluation points x
	   fy
	   fxx  (optional)
	   fyy  (optional)
	"""
	
	pAlpha, qAlpha = rbfRationalCoreMultiple(B,Bi,L,fc)
	Bp, Bq = B@pAlpha, B@qAlpha
	Hp, Hq = Hx@pAlpha, Hx@qAlpha
	fx = ( Bq*Hp - Bp*Hq )/Bq**2
	Hpy, Hqy = Hy@pAlpha, Hy@qAlpha
	fy = ( Bq*Hpy - Bp*Hqy )/Bq**2
	
	if second:
		Hpp, Hqq = H2x@pAlpha, H2x@qAlpha
		Hppy, Hqqy = H2y@pAlpha, H2y@qAlpha
		fxx = ( 2*Bp*Hq**2 + Bq**2*Hpp - Bq*( 2*Hp*Hq + Bp*Hqq ) )/Bq**3
		fyy = ( 2*Bp*Hqy**2 + Bq**2*Hppy - Bq*( 2*Hpy*Hqy + Bp*Hqqy ) )/Bq**3
		
	return fx, fy, fxx, fyy
	

# --------------------------------------------------------------------------
